/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 * Copyright (C) 2009 Canonical Services Ltd (www.canonical.com)
 *
 * Authors: Rodrigo Moya <rodrigo.moya@canonical.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of version 2 of the GNU Lesser General Public
 * License as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */

#include "config.h"
#include <webkit/webkit.h>
#include <dbus/dbus-glib.h>
#include <gnome-keyring.h>
#include "oauth.h"
#include "u1-music-store.h"
#include "u1-marshal.h"

#define U1_STORE_URL               "/music/login"
#define U1_NOT_LOGGED_IN_STORE_URL "/music/store-no-token"
#define U1_NOT_REGISTERED_URL      "/music/notregistered?return_url="

#define U1_DEFAULT_ERROR_PAGE "<html><body>Could not load Music Store</body></html>"
#define U1_IN_DEVELOPMENT_PAGE "<html><body>The music store is under development, please be patient</body></html>"

struct _U1MusicStorePrivate {
	gchar *base_url;

	GtkWidget *web_viewer;

	GdkCursor *busy_cursor;
	GdkCursor *normal_cursor;
};

G_DEFINE_TYPE(U1MusicStore, u1_music_store, GTK_TYPE_VBOX)

enum {
	PREVIEW_MP3_SIGNAL,
	LAST_SIGNAL
};

static guint u1_music_store_signals[LAST_SIGNAL] = { 0, };

static void
u1_music_store_finalize (GObject *object)
{
	U1MusicStore *music_store = U1_MUSIC_STORE (object);

	if (music_store->priv != NULL) {
		if (music_store->priv->base_url != NULL) {
			g_free (music_store->priv->base_url);
			music_store->priv->base_url;
		}

		g_free (music_store->priv);
		music_store->priv = NULL;
	}

	G_OBJECT_CLASS (u1_music_store_parent_class)->finalize (object);
}

static void
u1_music_store_class_init (U1MusicStoreClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS (klass);

	/* Signals */
	u1_music_store_signals[PREVIEW_MP3_SIGNAL] = g_signal_new ("preview-mp3",
								   G_TYPE_FROM_CLASS (klass),
								   (GSignalFlags) G_SIGNAL_RUN_LAST,
								   G_STRUCT_OFFSET (U1MusicStoreClass, preview_mp3),
								   NULL,
								   NULL,
								   _u1_marshal_VOID__STRING_STRING,
								   G_TYPE_NONE, 2,
								   G_TYPE_STRING, G_TYPE_STRING);
	object_class->finalize = u1_music_store_finalize;
}

static gboolean
navigation_requested_cb (WebKitWebView *web_view,
			 WebKitWebFrame *frame,
			 WebKitNetworkRequest *request,
			 WebKitWebNavigationAction *navigation_action,
			 WebKitWebPolicyDecision *policy_decision,
			 gpointer user_data)
{
	U1MusicStore *music_store = U1_MUSIC_STORE (user_data);

	/* Change the cursor when loading */
	if (music_store->priv->normal_cursor)
		gdk_cursor_unref (music_store->priv->normal_cursor);
#if GTK_CHECK_VERSION(2, 18, 0)
	music_store->priv->normal_cursor = gdk_window_get_cursor (GTK_WIDGET (web_view)->window);
#endif
	gdk_cursor_ref (music_store->priv->normal_cursor);

	if (music_store->priv->busy_cursor == NULL)
		music_store->priv->busy_cursor = gdk_cursor_new_for_display (gdk_display_get_default (), GDK_WATCH);

	gdk_window_set_cursor (GTK_WIDGET (web_view)->window, music_store->priv->busy_cursor);

	/* Inform webkit to continue with the loading */
	webkit_web_policy_decision_use (policy_decision);

	return TRUE;
}

static void
execute_script (WebKitWebView *web_view, const gchar *script_name, const gchar *replacement)
{
	gchar *file_contents, *script, *path;
	gsize length;
	GError *error = NULL;

	path = g_build_path ("/", U1_JAVASCRIPT_DIR, script_name, NULL);

	/* Load script from known location */
	if (g_file_get_contents (path, &file_contents, &length, &error)) {
		if (replacement != NULL)
			script = g_strdup_printf (file_contents, replacement);
		else
			script = g_strdup (file_contents);

		webkit_web_view_execute_script (web_view, script);

		g_free (script);
		g_free (file_contents);
	} else {
		g_warning ("Could not load %s: %s", path, error->message);
		g_error_free (error);
	}

	g_free (path);
}

static void
parse_oauth_string (const gchar *string, gchar **oauth_token, gchar **oauth_token_secret)
{
	gchar **items;

	*oauth_token = NULL;
	*oauth_token_secret = NULL;

	items = g_strsplit (string, "&", 2);
	if (items) {
		gint i;

		for (i = 0; items[i] != NULL; i++) {
			gchar **kv_pair;

			kv_pair = g_strsplit (items[i], "=", 2);
			if (kv_pair) {
				if (g_strcmp0 ((const gchar *) kv_pair[0], "oauth_token") == 0)
					*oauth_token = g_strdup (kv_pair[1]);
				else if (g_strcmp0 ((const gchar *) kv_pair[0], "oauth_token_secret") == 0)
					*oauth_token_secret = g_strdup (kv_pair[1]);

				g_strfreev (kv_pair);
			}
		}

		g_strfreev (items);
	}
}

static void
get_credentials_from_keyring (gchar **oauth_token, gchar **oauth_token_secret)
{
	GnomeKeyringAttributeList *attrs;
	GnomeKeyringResult result;
	GList *items_found;

	*oauth_token = *oauth_token_secret = NULL;

	/* Get the OAuth token from the keyring */
	attrs = gnome_keyring_attribute_list_new ();
	gnome_keyring_attribute_list_append_string (attrs, "ubuntuone-realm", "https://ubuntuone.com");
	gnome_keyring_attribute_list_append_string (attrs, "oauth-consumer-key", "ubuntuone");

	result = gnome_keyring_find_items_sync (GNOME_KEYRING_ITEM_GENERIC_SECRET,
						attrs, &items_found);
	if (result == GNOME_KEYRING_RESULT_OK && items_found != NULL) {
		GnomeKeyringFound *first_item = (GnomeKeyringFound *) items_found->data;
					
		parse_oauth_string (first_item->secret, oauth_token, oauth_token_secret);
		gnome_keyring_found_list_free (items_found);
	}
}

typedef struct {
	U1MusicStore *music_store;
	gchar *not_registered_url;
} NotRegisteredReplacementData;

static void
got_new_dbus_credentials_cb (DBusGProxy *proxy, const gchar *realm, const gchar *consumer_key, gpointer user_data)
{
	gchar *oauth_token, *oauth_token_secret, *url_to_use, *real_url;
	gchar **uri_parts;
	NotRegisteredReplacementData *nrrd = (NotRegisteredReplacementData *) user_data;

	get_credentials_from_keyring (&oauth_token, &oauth_token_secret);

	/* Get the return_url parameter from the not-registered url */
	uri_parts = g_strsplit (nrrd->not_registered_url, "?", 0);
	if (uri_parts != NULL && uri_parts[1] != NULL) {
		gchar **uri_args;

		uri_args = g_strsplit (uri_parts[1], "?", 0);
		if (uri_args != NULL) {
			gint i;

			for (i = 0; uri_args[i] != NULL; i++) {
				if (g_str_has_prefix (uri_args[i], "return_url")) {
					gchar **this_arg;

					this_arg = g_strsplit (uri_args[i], "=", 2);
					url_to_use = g_strdup_printf ("%s%s?return_url=%s",
								      nrrd->music_store->priv->base_url,
								      U1_STORE_URL,
								      this_arg[1]);

					g_strfreev (this_arg);
					break;
				}
			}

			g_strfreev (uri_args);
		}

		g_strfreev (uri_parts);
	} else
		url_to_use = g_strdup_printf ("%s%s", nrrd->music_store->priv->base_url, U1_STORE_URL);

	/* Now sign the URL and load the new page in the view */
	real_url = oauth_sign_url2 (url_to_use, NULL, OA_HMAC, "GET",
				    "ubuntuone", "hammertime",
				    oauth_token, oauth_token_secret);
	webkit_web_view_open (WEBKIT_WEB_VIEW (nrrd->music_store->priv->web_viewer), real_url);

	/* Free memory */
	g_object_unref (G_OBJECT (proxy));
	g_free (url_to_use);
	g_free (real_url);
	g_free (oauth_token);
	g_free (oauth_token_secret);

	g_free (nrrd->not_registered_url);
	g_free (nrrd);
}

static void
load_finished_cb (WebKitWebView *web_view, WebKitWebFrame *frame, gpointer user_data)
{
	const gchar *library_override;
	gchar *not_registered_url;
	U1MusicStore *music_store = U1_MUSIC_STORE (user_data);

	/* Inject some JavaScript to enable previews and hide the Flash player */
	execute_script (web_view, "u1-preview.js", NULL);

	/* Allow us to specifically override the library page links */
	library_override = g_getenv ("U1MUSICLIBRARYURL");
	if (library_override != NULL)
		execute_script (web_view, "u1-library-override.js", library_override);

	gdk_window_set_cursor (GTK_WIDGET (web_view)->window, music_store->priv->normal_cursor);

	/* If the URL just loaded is for not registered users, fire the OAuth authentication */
	not_registered_url = g_strdup_printf ("%s%s", music_store->priv->base_url, U1_NOT_REGISTERED_URL);
	if (g_str_has_prefix (webkit_web_view_get_uri (web_view), not_registered_url)) {
		DBusGConnection *bus;
		DBusGProxy *proxy;
		gboolean success;
		GError *error = NULL;

		bus = dbus_g_bus_get (DBUS_BUS_SESSION, &error);
		if (error) {
			g_warning ("Couldn't get session bus: %s", error->message);
			g_error_free (error);

			return;
		}

		proxy = dbus_g_proxy_new_for_name (bus, "com.ubuntuone.Authentication",
						   "/", "com.ubuntuone.Authentication");
		if (proxy != NULL) {
			NotRegisteredReplacementData *nrrd;

			nrrd = g_new0 (NotRegisteredReplacementData, 1);
			nrrd->music_store = music_store;
			nrrd->not_registered_url = g_strdup (webkit_web_view_get_uri (web_view));

			dbus_g_proxy_add_signal (proxy, "NewCredentials", G_TYPE_STRING, G_TYPE_STRING, G_TYPE_INVALID);
			dbus_g_proxy_connect_signal (proxy, "NewCredentials",
						     (GCallback) got_new_dbus_credentials_cb,
						     nrrd,
						     NULL);

			error = NULL;
			success = dbus_g_proxy_call (proxy, "login", &error,
						     G_TYPE_INVALID,
						     G_TYPE_STRING, "ubuntuone",
						     G_TYPE_STRING, "hammertime",
						     G_TYPE_INVALID);

			/* Free memory */
			dbus_g_connection_unref (bus);
		} else {
			g_warning ("Couldn't get proxy for com.ubuntuone.Authentication");
			return;

		}
	}

	g_free (not_registered_url);
}

static void
status_bar_text_changed_cb (WebKitWebView *web_view, const gchar *text, gpointer user_data)
{
	U1MusicStore *music_store = U1_MUSIC_STORE (user_data);

	if (g_str_has_prefix (text, "u1preview:::")) {
		gchar **args;

		args = g_strsplit (text, ":::", 0);
		if (args != NULL) {
			g_signal_emit (music_store, u1_music_store_signals[PREVIEW_MP3_SIGNAL], 0, args[1], args[2]);
			g_strfreev (args);
		}
	}
}

static gboolean
load_error_cb (WebKitWebView *web_view, WebKitWebFrame *frame, const gchar *uri, GError *error, gpointer user_data)
{
	U1MusicStore *music_store = U1_MUSIC_STORE (user_data);

	webkit_web_view_load_string (web_view, U1_DEFAULT_ERROR_PAGE, "text/html", "utf-8", "file:///");

	return TRUE;
}

static gchar *
get_url_to_use (U1MusicStore *music_store)
{
	gchar *oauth_token = NULL, *oauth_token_secret = NULL, *real_url;
	const gchar *url_to_use;

	/* If U1MUSICSTOREURL is defined, use that instead of the real URL */
	if (! (url_to_use = g_getenv ("U1MUSICSTOREURL")))
		url_to_use = "https://one.ubuntu.com";

	if (music_store->priv->base_url != NULL)
		g_free (music_store->priv->base_url);

	music_store->priv->base_url = g_strdup (url_to_use);

	/* OAUth sign the URL */
	if (g_str_has_prefix (url_to_use, "http://localhost") || g_str_has_prefix (url_to_use, "http://127.0.0.1")) {
		const gchar *oauthfile;

		oauthfile = g_getenv ("OAUTHKEYFILE");
		if (oauthfile != NULL) {
			gchar *oauthstring;
			gsize len;

			if (g_file_get_contents (oauthfile, &oauthstring, &len, NULL)) {
				parse_oauth_string (oauthstring, &oauth_token, &oauth_token_secret);
				g_free (oauthstring);
			}
		} else {
			webkit_web_view_load_string (WEBKIT_WEB_VIEW (music_store->priv->web_viewer),
						     "Fail! If you specify U1MUSICSTOREURL you must "
						     "also specify OAUTHKEYFILE as a file with OAuth keys in it",
						     "text/html", "utf-8", "file:///");
			return NULL;
		}
	} else
		get_credentials_from_keyring (&oauth_token, &oauth_token_secret);

	if (oauth_token == NULL || oauth_token_secret == NULL)
		real_url = g_strdup_printf ("%s%s", music_store->priv->base_url, U1_NOT_LOGGED_IN_STORE_URL);
	else {
		gchar *s = g_strdup_printf ("%s%s", music_store->priv->base_url, url_to_use);

		real_url = oauth_sign_url2 (s, NULL, OA_HMAC, "GET",
					    "ubuntuone", "hammertime",
					    oauth_token, oauth_token_secret);
		g_free (s);
	}

	g_free (oauth_token);
	g_free (oauth_token_secret);

	return real_url;
}

static void
u1_music_store_init (U1MusicStore *music_store)
{
	GtkWidget *scroll;

	music_store->priv = g_new0 (U1MusicStorePrivate, 1);

	/* Create web viewer object */
	scroll = gtk_scrolled_window_new (NULL, NULL);
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scroll),
					GTK_POLICY_AUTOMATIC,
					GTK_POLICY_AUTOMATIC);
	gtk_widget_show (scroll);

	music_store->priv->web_viewer = webkit_web_view_new ();
	g_object_set (G_OBJECT (webkit_web_view_get_settings (WEBKIT_WEB_VIEW (music_store->priv->web_viewer))),
		      "user-agent", "U1MusicStore widget (" VERSION ")",
		      NULL);

	g_signal_connect (G_OBJECT (music_store->priv->web_viewer), "navigation-policy-decision-requested",
			  G_CALLBACK (navigation_requested_cb), music_store);
	g_signal_connect (G_OBJECT (music_store->priv->web_viewer), "load-finished",
			  G_CALLBACK (load_finished_cb), music_store);
	g_signal_connect (G_OBJECT (music_store->priv->web_viewer), "status-bar-text-changed",
			  G_CALLBACK (status_bar_text_changed_cb), music_store);
	g_signal_connect (G_OBJECT (music_store->priv->web_viewer), "load-error",
			  G_CALLBACK (load_error_cb), music_store);
	gtk_widget_show (music_store->priv->web_viewer);
	gtk_container_add (GTK_CONTAINER (scroll), music_store->priv->web_viewer);

	/* FIXME: music store is broken right now, so prevent people from using it */
	if (g_strcmp0 (g_getenv ("THEMUSICSTOREISBROKEN"), "yes") != 0)
		webkit_web_view_load_string (WEBKIT_WEB_VIEW (music_store->priv->web_viewer),
					     U1_IN_DEVELOPMENT_PAGE, "text/html", "utf-8", "file:///");
	else {
		gchar *real_url;

		real_url = get_url_to_use (music_store);
		if (real_url != NULL) {
			webkit_web_view_open (WEBKIT_WEB_VIEW (music_store->priv->web_viewer), real_url);

			/* Free memory */
			g_free (real_url);
		}
	}

	gtk_box_pack_start (GTK_BOX (music_store), scroll, TRUE, TRUE, 3);
}

/**
 * u1_music_store_new:
 *
 * Create a new #U1MusicStore widget.
 *
 * Return value: the newly created widget.
 */
GtkWidget *
u1_music_store_new (void)
{
	U1MusicStore *music_store;

	music_store = g_object_new (U1_TYPE_MUSIC_STORE, NULL);

	return (GtkWidget *) music_store;
}
